package ru.dbotthepony.mc.otm.item

import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.formatPower
import ru.dbotthepony.mc.otm.saveddata.SavedCountingMap
import ru.dbotthepony.mc.otm.saveddata.SavedMapDelegate

class QuantumBatteryItem : Item {
	private inner class Power : IMatteryEnergyStorage, ICapabilityProvider, INBTSerializable<IntTag> {
		private val resolver = LazyOptional.of { this }
		var delegate = SavedMapDelegate(ImpreciseFraction.ZERO)

		override fun <T : Any?> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
			if (cap == ForgeCapabilities.ENERGY || cap == MatteryCapability.ENERGY) {
				return resolver.cast()
			}

			return LazyOptional.empty()
		}

		override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (howMuch.isNegative) {
				return ImpreciseFraction.ZERO
			}

			if (isCreative) {
				val newEnergy = (delegate.value - howMuch).moreThanZero()

				if (!simulate) {
					delegate.value = newEnergy
				}

				return delegate.value - newEnergy
			}

			val newEnergy = (delegate.value - howMuch.coerceAtMost(throughput!!)).moreThanZero()

			if (!simulate) {
				delegate.value = newEnergy
			}

			return delegate.value - newEnergy
		}

		override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (howMuch.isNegative) {
				return ImpreciseFraction.ZERO
			}

			if (isCreative) {
				if (!simulate) {
					delegate.value += howMuch
				}

				return howMuch
			}

			val newEnergy = (delegate.value + howMuch.coerceAtMost(throughput!!)).coerceAtMost(capacity!!)

			if (!simulate) {
				delegate.value = newEnergy
			}

			return newEnergy - delegate.value
		}

		override val batteryLevel: ImpreciseFraction
			get() = delegate.value

		override val maxBatteryLevel: ImpreciseFraction
			get() = capacity ?: (batteryLevel + ImpreciseFraction.LONG_MAX_VALUE)

		override val missingPower: ImpreciseFraction
			get() = if (isCreative) ImpreciseFraction.LONG_MAX_VALUE else super.missingPower

		override fun canExtract(): Boolean {
			return true
		}

		override fun canReceive(): Boolean {
			return true
		}

		override fun serializeNBT(): IntTag {
			if (isServerThread() && delegate.parent == null) {
				val old = delegate
				delegate = saveData!!.factorize()
				delegate.value = old.value
			}

			return IntTag.valueOf(if (delegate.parent == null) Int.MIN_VALUE else delegate.index)
		}

		override fun deserializeNBT(nbt: IntTag) {
			if (nbt.asInt == Int.MIN_VALUE) {
				delegate = SavedMapDelegate(ImpreciseFraction.ZERO)
				return
			}

			delegate = saveData?.computeIfAbsent(nbt.asInt) ?: SavedMapDelegate(null, nbt.asInt, delegate.value)
		}
	}

	val isCreative: Boolean
	val capacity: ImpreciseFraction?
	val throughput: ImpreciseFraction?

	val saveDataID: String

	val saveData: SavedCountingMap<SavedMapDelegate<ImpreciseFraction>>? get() {
		if (isServerThread()) {
			return MINECRAFT_SERVER.overworld().dataStorage.computeIfAbsent({
				SavedMapDelegate.makeMap(ImpreciseFraction::serializeNBT, ImpreciseFraction.Companion::deserializeNBT, ImpreciseFraction.ZERO).load(it)
			}, {
				SavedMapDelegate.makeMap(ImpreciseFraction::serializeNBT, ImpreciseFraction.Companion::deserializeNBT, ImpreciseFraction.ZERO)
			}, saveDataID) ?: throw NullPointerException("Unable to get save data for $this in ${MINECRAFT_SERVER.overworld()}")
		}

		return null
	}

	constructor(saveDataID: String) : super(Properties().stacksTo(1).rarity(Rarity.EPIC).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
		isCreative = true
		capacity = null
		throughput = null
		this.saveDataID = "otm_$saveDataID".intern()
	}

	constructor(saveDataID: String, capacity: ImpreciseFraction, io: ImpreciseFraction) : super(Properties().stacksTo(1).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
		isCreative = false
		this.capacity = capacity
		this.throughput = io
		this.saveDataID = "otm_$saveDataID".intern()
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return Power()
	}

	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		components: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, components, p_41424_)

		val power = itemStack.getCapability(MatteryCapability.ENERGY).orThrow() as Power

		components.add(TranslatableComponent("otm.item.quantum_description").withStyle(ChatFormatting.DARK_GRAY))

		if (isCreative) {
			components.add(TranslatableComponent("otm.item.quantum_battery.creative_power", power.batteryLevel.formatPower()).withStyle(ChatFormatting.GRAY))
			components.add(TranslatableComponent("otm.item.quantum_battery.creative").withStyle(ChatFormatting.DARK_GRAY))
			components.add(TranslatableComponent("otm.item.quantum_battery.creative2").withStyle(ChatFormatting.DARK_GRAY))
		} else {
			components.add(TranslatableComponent("otm.item.power.normal.storage", power.batteryLevel.formatPower(), capacity!!.formatPower()).withStyle(ChatFormatting.GRAY))
			components.add(TranslatableComponent(
				"otm.item.power.normal.throughput",
				throughput!!.formatPower(),
				throughput.formatPower()
			).withStyle(ChatFormatting.GRAY))
		}

		components.add(TranslatableComponent("otm.item.quantum_link_id", power.delegate.index).withStyle(ChatFormatting.DARK_GRAY))
	}
}
